--[[
  Pano professional plugin for darktable 2.2.X,2.4.X,2.6.X

  copyright (c) 2017, 2018  Holger Klemm
  copyright (c) 2018, Andreas Urbschat 

  darktable is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  darktable is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with darktable.  If not, see <http://www.gnu.org/licenses/>.
]]

--[[
Version 2.1.3 for darktable 2.2.X,2.4.X,2.6.X

Depends:
- pto_gen
- cpfind
- cpclean
- linefind
- autooptimiser
- pano_modify
- nona
- enblend 4.2
- exiftool
- darktable lua lib (https://github.com/darktable-org/lua-scripts)

Known bugs:
- nona doesn't work with 32bit TIFF inputfiles. please use the 16bit or 8bit TIFF image export

Release note:
- Add support for Windows (from Andreas Urbschat)
- the script now requires the library from the standard darktable lua scripts
- without directory, the images are written to the source directory

Info workflow:
pto_gen -o project.pto *.jpg
cpfind -o project.pto --multirow --celeste project.pto
cpclean -o project.pto project.pto
linefind -o project.pto project.pto
autooptimiser -a -m -l -s -o project.pto project.pto
pano_modify --canvas=AUTO --crop=AUTO -o project.pto project.pto
nona -m TIFF_m -o project project.pto
enblend -o project.tif project0000.tif project0001.tif project0002.tif
exiftool

ADDITIONAL SOFTWARE NEEDED FOR THIS SCRIPT


USAGE
* require this file from your main luarc config file.

This plugin will add the new export modul "panorama image".
]]

local dt = require "darktable"
local df = require "lib/dtutils.file"
local log = require "lib/dtutils.log"
local gettext = dt.gettext

local namespace = 'pano_pro'

local executable_table = {"pto_gen", "cpfind", "cpclean", "linefind", "autooptimiser", "pano_modify", "nona", "enblend", "exiftool"}


-- works with LUA API version 4.0.0
dt.configuration.check_version(...,{4,0,0},{5,0,0})
dt.print_error("Pano Pro plugin version 2.1.3 loaded") 

-- Tell gettext where to find the .mo file translating messages for a particular domain
gettext.bindtextdomain(namespace, dt.configuration.config_dir.."/lua/")

local function _(msgid)
    return gettext.dgettext(namespace, msgid)
end


-- command variables
local pano_pro_tmppath=dt.configuration.tmp_dir
local pano_pro_images_to_process =""
local pano_pro_first_imagefile =""
local pano_pro_first_last_imagefile=""
local pano_pro_counted_images=0
local pano_pro_cmd_output_path=""
local pano_pro_cmd_output_image_index=0
local pano_pro_path_with_filename=""

-- pto_gen
local pano_pro_cmd_pto_gen_projection_type=""
local pano_pro_cmd_pto_gen_distortion=""
local pano_pro_cmd_pto_gen_vignetting=""
local pano_pro_cmd_pto_gen_fov=""
local pano_pro_cmd_pto_gen_output=""
local pano_pro_pto_genStartCommand=""
-- cpfind
local pano_pro_pto_cmd_cpfind_matching=""
local pano_pro_cmd_cpfind_celeste=""
local pano_pro_cmd_cpfind_celeste_radius=""
local pano_pro_cmd_cpfind_celeste_threshold=""
local pano_pro_cmd_cpfind_ransacmode=""
local pano_pro_cmd_cpfind_ransaciter=""
local pano_pro_cmd_cpfind_ransacdist=""
local pano_pro_cmd_cpfind_cache=""
local pano_pro_cmd_cpfind_fullscale=""
local pano_pro_cmd_cpfind_input=""
local pano_pro_cmd_cpfind_output=""
local pano_pro_cpfindStartCommand=""
-- cpclean
local pano_pro_cmd_cpclean_optimise_whole_panorama=""
local pano_pro_cmd_cpclean_input=""
local pano_pro_cmd_cpclean_output=""
local pano_pro_cpcleanStartCommand=""
-- linefind
local pano_pro_cmd_linefind_input=""
local pano_pro_cmd_linefind_output=""
local pano_pro_linefindStartCommand=""
--autooptimiser
local pano_pro_cmd_autooptimiser_auto_align_mode=""
local pano_pro_cmd_autooptimiser_pairwise_optimisation=""
local pano_pro_cmd_autooptimiser_photometric_parameters=""
local pano_pro_cmd_autooptimiser_level_horizon=""
local pano_pro_cmd_autooptimiser_auto_projection=""
local pano_pro_cmd_autooptimiser_input=""
local pano_pro_cmd_autooptimiser_output=""
local pano_pro_autooptimiserStartCommand=""
--pano_modify
local pano_pro_cmd_pano_modify_canvas=""
local pano_pro_cmd_pano_modify_crop=""
local pano_pro_cmd_pano_modify_straighten=""
local pano_pro_cmd_pano_modify_center=""
local pano_pro_cmd_pano_modify_exposure=""
local pano_pro_cmd_pano_modify_input=""
local pano_pro_cmd_pano_modify_output=""
local pano_pro_cmd_pano_modify_output_format=""
local pano_pro_cmd_pano_modify_output_compression=""
local pano_pro_pano_modifyStartCommand=""
--nona
local pano_pro_cmd_nona_exposure=""
local pano_pro_cmd_nona_output_format=""
local pano_pro_cmd_nona_output_depth=""
local pano_pro_cmd_nona_output_compression=""
local pano_pro_cmd_nona_use_gpu=""
local pano_pro_cmd_nona_prj_input=""
local pano_pro_cmd_nona_bitmap_output=""
local pano_pro_nonaStartCommand=""
--enblend
local pano_pro_cmd_enblend_output_format=""
local pano_pro_cmd_enblend_output_depth=""
local pano_pro_cmd_enblend_output_compression=""
local pano_pro_cmd_enblend_use_gpu=""
local pano_pro_cmd_enblend_bitmap_input=""
local pano_pro_cmd_enblend_bitmap_output=""
local pano_pro_enblendStartCommand=""

-- initialize combobox and checkbox value for default
if (dt.preferences.read(namespace,  "selected_projection_type", "integer") == 0) then
-- pto_gen defaults
  log.msg(log.debug, "initialize combobox and checkbox value for default")
  dt.preferences.write(namespace, "selected_projection_type", "integer", 3)
  dt.preferences.write(namespace, "checked_distortion", "bool", true)
  dt.preferences.write(namespace, "checked_vignetting", "bool", true)
  dt.preferences.write(namespace, "checked_fov", "bool", false)
  dt.preferences.write(namespace, "set_fov", "float", 50.0)
  dt.preferences.write(namespace, "visible_fov", "bool", false)
  dt.preferences.write(namespace, "selected_matching_strategy", "integer", 1)
  dt.preferences.write(namespace, "checked_celeste", "bool", true)
  dt.preferences.write(namespace, "selected_celeste_radius", "integer", 11)
  dt.preferences.write(namespace, "visible_celeste_radius", "bool", true)
  dt.preferences.write(namespace, "selected_celeste_threshold", "integer", 5)
  dt.preferences.write(namespace, "visible_celeste_threshold", "bool", true)
  dt.preferences.write(namespace, "checked_fullscale", "bool", false)
  dt.preferences.write(namespace, "checked_whole_pano_checking", "bool", true)
  dt.preferences.write(namespace, "selected_ransacmode", "integer", 1)
  dt.preferences.write(namespace, "selected_ransac_iterations", "integer", 2)
  dt.preferences.write(namespace, "selected_ransac_distance", "integer", 11)
  dt.preferences.write(namespace, "checked_auto_align_mode", "bool", true)
  dt.preferences.write(namespace, "checked_pairwise_optimisation", "bool", false)
  dt.preferences.write(namespace, "checked_optimise_photometric_parameters", "bool", true)
  dt.preferences.write(namespace, "checked_level_horizon", "bool", true)
  dt.preferences.write(namespace, "checked_auto_projection_size", "bool", true)
  dt.preferences.write(namespace, "selected_canvas_size", "integer", 1)
  dt.preferences.write(namespace, "selected_output_crop", "integer", 1)
  dt.preferences.write(namespace, "checked_straighten", "bool", false)
  dt.preferences.write(namespace, "checked_center", "bool", true)
  dt.preferences.write(namespace, "selected_exposure", "integer", 1)
  dt.preferences.write(namespace, "set_exposure", "float", 2.0)
  dt.preferences.write(namespace, "visible_exposure", "bool", false)
  dt.preferences.write(namespace, "selected_output_format", "integer", 2)
  dt.preferences.write(namespace, "selected_overwrite", "integer", 1)
end

local pano_pro_label_project_options= dt.new_widget("label")
{
  label = _('project options'),
  ellipsize = "start",
  halign = "end"
}

local pano_pro_label_line1= dt.new_widget("label")
{
  label = "_______________________________________________",
  ellipsize = "start",
  halign = "end"
}


local pano_pro_combobox_projection_type = dt.new_widget("combobox")
{
  label = _('projection type'),
  tooltip =_('Sets the output projection. For more informations have a look at:\nhttp://wiki.panotools.org/Projections'),
  value = dt.preferences.read(namespace, "selected_projection_type", "integer"), --0

  changed_callback = function(sel_projection_type)
    dt.preferences.write(namespace, "selected_projection_type", "integer", sel_projection_type.selected)
  end,
  "rectilinear", "panorama (circular)","equirectangular","fisheye_ff", "stereographic","mercator","trans_mercator","sinusoidal","lambert_equal_area_conic","lambert_azimuthal","albers_equal_conic","miller_cylindrical","panini","architectural","orthographic","equisolid","equi_panini","biplane","triplane","panini_general","thoby","hammer",
   reset_callback = function(self_projection_type)
     self_projection_type.value = 3
  end
}


local pano_pro_check_button_distortion = dt.new_widget("check_button")
{
    label = _('distortion correction'),
    value = dt.preferences.read(namespace, "checked_distortion", "bool"),
    tooltip =_('try to load distortion from database'),
    clicked_callback = function(distortion)
        if (distortion.value) then
           dt.preferences.write(namespace, "checked_distortion", "bool", true)
        else
           dt.preferences.write(namespace, "checked_distortion", "bool", false)
        end
    end,
    reset_callback = function(self_button_distortion)
       self_button_distortion.value = true
       dt.preferences.write(namespace, "checked_distortion", "bool", true)
    end
}


local pano_pro_check_button_vignetting = dt.new_widget("check_button")
{
    label = _('vignetting correction'),
    value = dt.preferences.read(namespace, "checked_vignetting", "bool"),
    tooltip =_('try to load vignetting from database'),
    clicked_callback = function(vignetting)
    if (vignetting.value) then
             dt.preferences.write(namespace, "checked_vignetting", "bool", true)
          else
             dt.preferences.write(namespace, "checked_vignetting", "bool", false)
          end
    end,
    reset_callback = function(self_button_vignetting)
       self_button_vignetting.value = true
       dt.preferences.write(namespace, "checked_vignetting", "bool", true)
    end
}


local pano_pro_slider_fov = dt.new_widget("slider")
{
  label = _('horizontal field of view'),
  tooltip = _('Sets the horizontal field of view for all images. Useful if your lens does not store the focal length and/or crop factor correctly in the EXIF data.'),
  hard_min = 5,
  hard_max = 200,
  soft_min = 5,
  soft_max = 200,
  value = dt.preferences.read(namespace, "set_fov", "float"), --50 default
  sensitive= dt.preferences.read(namespace, "visible_fov", "bool"),
}

local pano_pro_check_button_fov = dt.new_widget("check_button")
{
    label = _('horizontal field of view'),
    value = dt.preferences.read(namespace, "checked_fov", "bool"),
    tooltip =_('Sets the horizontal field of view for all images. Useful if your lens does not store the focal length and/or crop factor correctly in the EXIF data.'),
    clicked_callback = function(fov)
      if (fov.value) then
        dt.preferences.write(namespace, "checked_fov", "bool", true)
        dt.preferences.write(namespace, "visible_fov", "bool", true)
        pano_pro_slider_fov.sensitive=true
      else
        dt.preferences.write(namespace, "checked_fov", "bool", false)
        dt.preferences.write(namespace, "visible_fov", "bool", false)
        pano_pro_slider_fov.sensitive=false
      end
    end,
    reset_callback = function(self_button_fov)
       self_button_fov.value = false
       dt.preferences.write(namespace, "checked_fov", "bool", false)
       dt.preferences.write(namespace, "visible_fov", "bool", false)
        pano_pro_slider_fov.sensitive=false
    end
}



local pano_pro_separator1 = dt.new_widget("separator")
{
}

local pano_pro_label_control_points= dt.new_widget("label")
{
     label = _('control point options'),
     ellipsize = "start",
     halign = "end"
}

local pano_pro_label_line2= dt.new_widget("label")
{
     label = "_______________________________________________",
     ellipsize = "start",
     halign = "end"
}


local pano_pro_combobox_matching_strategy = dt.new_widget("combobox")
{
    label = _('matching strategy'),
    tooltip =_("all pairs:\nthis is the default matching strategy. here all image pairs are matched against each other. E.g. if your project contains 5 images then cpfind matches the image pairs: 0-1, 0-2, 0-3, 0-4, 1-2, 1-3, 1-4, 2-3, 2-4 and 3-4\n\nlinear match:\nthis will only detect matches between adjacent images, e.g. for the 5 image example it will matches images pairs 0-1, 1-2, 2-3 and 3-4\n\nmultirow matching:\nThe algorithm is the same as described in multi-row panorama. By integrating this algorithm intocpfind it is faster by using several cores of modern CPUs and don't caching the keypoints to disc (which is time consuming)."),
    value = dt.preferences.read(namespace, "selected_matching_strategy", "integer"), --0
    changed_callback = function(sel_projection_type)
    end,
    "all pairs", "linear match", "multirow matching",
    reset_callback = function(self_points_error)
       self_points_error.value = dt.preferences.read(namespace, "selected_matching_strategy", "integer")
    end
}


local pano_pro_combobox_celeste_radius = dt.new_widget("combobox")
{
    label = _('celeste radius'),
    tooltip =_('the celeste step can be fine tuned by the parameters'),
    sensitive=dt.preferences.read(namespace, "visible_celeste_radius", "bool"),
    value = dt.preferences.read(namespace, "selected_celeste_radius", "integer"), --1
    changed_callback = function(sel_celeste_radius)
    end,
    "10", "11", "12","13","14","15","16","17","18","19","20","21","22","23","24","25","26","27","28","29","30",
    reset_callback = function(self_celeste_radius)
       self_celeste_radius.value = dt.preferences.read(namespace, "selected_celeste_radius", "integer")
    end
}


local pano_pro_combobox_celeste_threshold = dt.new_widget("combobox")
{
    label = _('celeste threshold'),
    tooltip =_('the celeste step can be fine tuned by the parameters'),
    sensitive = dt.preferences.read(namespace, "visible_celeste_threshold", "bool"),
    value = dt.preferences.read(namespace, "selected_celeste_threshold", "integer"), --1
    changed_callback = function(sel_celeste_threshold)
    end,
    "0.1", "0.2", "0.3","0.4","0.5","0.6","0.7","0.8","0.9","1.0",
    reset_callback = function(self_celeste_threshold)
       self_celeste_threshold.value = dt.preferences.read(namespace, "selected_celeste_threshold", "integer")
    end
}


local pano_pro_check_button_celeste = dt.new_widget("check_button")
{
    label = _('celeste algorithm'),
    value = dt.preferences.read(namespace, "checked_celeste", "bool"),
    tooltip =_('outdoor panorama often contains clouds. clouds are bad areas for setting control points because they are moving object. the celeste algorithm can masked out areas which contains clouds. this is only done internal for the keypoint finding step and does not change the alpha channel of your image.'),
    clicked_callback = function(celeste)
    if (celeste.value) then
             dt.preferences.write(namespace, "checked_celeste", "bool", true)
             dt.preferences.write(namespace, "visible_celeste_radius", "bool", true)
             dt.preferences.write(namespace, "visible_celeste_threshold", "bool", true)
             pano_pro_combobox_celeste_radius.sensitive=true
             pano_pro_combobox_celeste_threshold.sensitive=true
          else
             dt.preferences.write(namespace, "checked_celeste", "bool", false)
             dt.preferences.write(namespace, "visible_celeste_radius", "bool", false)
             dt.preferences.write(namespace, "visible_celeste_threshold", "bool", false)
             pano_pro_combobox_celeste_radius.sensitive=false
             pano_pro_combobox_celeste_threshold.sensitive=false
          end
    end,
    reset_callback = function(self_celeste)
       self_celeste.value = true
       pano_pro_combobox_celeste_radius.sensitive=true
       pano_pro_combobox_celeste_threshold.sensitive=true
       dt.preferences.write(namespace, "checked_celeste", "bool", true)
       dt.preferences.write(namespace, "visible_celeste_radius", "bool", true)
       dt.preferences.write(namespace, "visible_celeste_threshold", "bool", true)
    end
}




local pano_pro_check_button_whole_pano_checking = dt.new_widget("check_button")
{
    label = _('pairwise optimisation from control points'),
    value = dt.preferences.read(namespace, "checked_whole_pano_checking", "bool"),
    tooltip =_('optimises all images pairs, calculates for each pair mean and standard deviation and removes all control points with error bigger than mean+n*sigma. this option skips the optimisation for the whole panorama'),
    clicked_callback = function(wholepanochecking)
    if (wholepanochecking.value) then
             dt.preferences.write(namespace, "checked_whole_pano_checking", "bool", true)
          else
             dt.preferences.write(namespace, "checked_whole_pano_checking", "bool", false)
          end
    end,
    reset_callback = function(self_whole_pano_checking)
      self_whole_pano_checking.value=true
      dt.preferences.write(namespace, "checked_whole_pano_checking", "bool", true)
    end
}


local pano_pro_check_button_fullscale= dt.new_widget("check_button")
{
    label = _('uses full scale image to detect keypoints'),
    value = dt.preferences.read(namespace, "checked_fullscale", "bool"),
    tooltip =_('For speed reasons cpfind is using images, which are scaled to their half width and height, to find keypoints. With this option cpfind is working on the full scale images. This takes longer but can provide \"better\" and/or more control points.'),
    clicked_callback = function(fullscale)
    if (fullscale.value) then
             dt.preferences.write(namespace, "checked_fullscale", "bool", true)
          else
             dt.preferences.write(namespace, "checked_fullscale", "bool", false)
          end
    end,
    reset_callback = function(self_fullscale)
      self_fullscale.value=true
      dt.preferences.write(namespace, "checked_fullscale", "bool", true)
    end
}




local pano_pro_combobox_ransacmode = dt.new_widget("combobox")
{
    label = _('random sample consensus mode'),
    tooltip =_('auto:\nuse homography for images with hfov < 65 degrees and rpy otherwise.\n\nhom:\nAssume a homography. Only applicable for non-wide angle views. Uses the original panomatic code. It is also more flexible than required and can generate false matches, particularly if most of the matches are located on a single line.\n\nrpy:\nAlign images using roll, pitch and yaw. This requires a good estimate for the horizontal field of view (and distortion, for heavily distorted images). It is the preferred mode if a calibrated lens is used, or the HFOV could be read successfully from the EXIF data.\n\nrpyv:\nAlign pair by optimizing roll, pitch, yaw and field of view. Should work without prior knowledge of the field of view, but might fail more often, due to error function used in the panotools optimizer, it tends to shrink the fov to 0.\n\nrpyvb:\nAlign pair by optimizing roll, pitch, yaw, field of view and the "b" distortion parameter.  Probably very fragile, just implemented for testing.'),
    value = dt.preferences.read(namespace, "selected_ransacmode", "integer"), --1
    changed_callback = function(sel_ransacmode)
    end,
    "auto", "hom", "rpy","rpyv","rpyb",
    reset_callback = function(self_ransacmode)
       self_ransacmode.value = 1
       dt.preferences.write(namespace, "selected_ransacmode", "integer", 1)
    end
}

local pano_pro_combobox_ransac_iterations = dt.new_widget("combobox")
{
    label = _('random sample consensus iterations'),
    tooltip =_('ransac: iterations (default : 1000)'),
    value = dt.preferences.read(namespace, "selected_ransac_iterations", "integer"), --1
    changed_callback = function(sel_ransaciter)
    end,
    "500", "1000", "1500","2000","2500",
    reset_callback = function(self_sel_ransaciter)
       self_sel_ransaciter.value = 2
       dt.preferences.write(namespace, "selected_ransac_iterations", "integer", 2)
    end
}

local pano_pro_combobox_ransac_distance = dt.new_widget("combobox")
{
    label = _('random sample consensus distance'),
    tooltip =_('ransac: homography estimation distance threshold (pixels) (default : 25)'),
    value = dt.preferences.read(namespace, "selected_ransac_distance", "integer"), --1
    changed_callback = function(sel_ransacdist)
    end,
    "15", "16", "17","18","19","20","21","22","23","24","25","26","27","28","29","30","31","32","33","34","35",
    reset_callback = function(self_ransacdis)
       self_ransacdis.value = 11
       dt.preferences.write(namespace, "selected_ransac_distance", "integer", 11)
    end
}

local pano_pro_separator2 = dt.new_widget("separator")
{
}

local pano_pro_label_pano_options= dt.new_widget("label")
{
     label = _('panorama options'),
     ellipsize = "start",
     halign = "end"
}

local pano_pro_label_line3= dt.new_widget("label")
{
     label = "_______________________________________________",
     ellipsize = "start",
     halign = "end"
}


local pano_pro_check_button_alignmode = dt.new_widget("check_button")
{
    label = _('auto align mode'),
    value = dt.preferences.read(namespace, "checked_auto_align_mode", "bool"),
    tooltip =_('auto align mode, includes various optimisation stages, depending on the amount and distribution of the control points'),
    clicked_callback = function(auto_align_mode)
    if (auto_align_mode.value) then
             dt.preferences.write(namespace, "checked_auto_align_mode", "bool", true)
          else
             dt.preferences.write(namespace, "checked_auto_align_mode", "bool", false)
          end
    end,
    reset_callback = function(self_alignmode)
      self_alignmode.value=true
      dt.preferences.write(namespace, "checked_auto_align_mode", "bool", true)
    end
}


local pano_pro_check_button_pairwise_optimisation = dt.new_widget("check_button")
{
    label = _('pairwise align optimisation'),
    value = dt.preferences.read(namespace, "checked_pairwise_optimisation", "bool"),
    tooltip =_('auto align mode, includes various optimisation stages, depending on the amount and distribution of the control points'),
    clicked_callback = function(pairwise_optimisation)
    if (pairwise_optimisation.value) then
             dt.preferences.write(namespace, "checked_pairwise_optimisation", "bool", true)
          else
             dt.preferences.write(namespace, "checked_pairwise_optimisation", "bool", false)
          end
    end,
    reset_callback = function(self_pairwise_optimisation)
      self_pairwise_optimisation.value=false
      dt.preferences.write(namespace, "checked_pairwise_optimisation", "bool", false)
    end
}


local pano_pro_check_button_optimise_photometric_parameters = dt.new_widget("check_button")
{
    label = _('optimise photometric parameters'),
    value = dt.preferences.read(namespace, "checked_optimise_photometric_parameters", "bool"),
    tooltip =_('optimise photometric parameters'),
    clicked_callback = function(optimise_photometric_parameters)
    if (optimise_photometric_parameters.value) then
             dt.preferences.write(namespace, "checked_optimise_photometric_parameters", "bool", true)
          else
             dt.preferences.write(namespace, "checked_optimise_photometric_parameters", "bool", false)
          end
    end,
    reset_callback = function(self_optimise_photometric_parameters)
      self_optimise_photometric_parameters.value=true
      dt.preferences.write(namespace, "checked_optimise_photometric_parameters", "bool", true)
    end
}


local pano_pro_check_button_level_horizon = dt.new_widget("check_button")
{
    label = _('level horizon'),
    value = dt.preferences.read(namespace, "checked_level_horizon", "bool"),
    tooltip =_('Level horizon (works best for horizontal panoramas)'),
    clicked_callback = function(level_horizon)
    if (level_horizon.value) then
             dt.preferences.write(namespace, "checked_level_horizon", "bool", true)
          else
             dt.preferences.write(namespace, "checked_level_horizon", "bool", false)
          end
    end,
    reset_callback = function(self_level_horizon)
      self_level_horizon.value=true
      dt.preferences.write(namespace, "checked_level_horizon", "bool", true)
    end
}


local pano_pro_check_button_auto_projection_size = dt.new_widget("check_button")
{
    label = _('auto output projection and size'),
    value = dt.preferences.read(namespace, "checked_auto_projection_size", "bool"),
    tooltip =_('automatically select a suitable output projection and size'),
    clicked_callback = function(auto_projection_size)
    if (auto_projection_size.value) then
             dt.preferences.write(namespace, "checked_auto_projection_size", "bool", true)
          else
             dt.preferences.write(namespace, "checked_auto_projection_size", "bool", false)
          end
    end,
    reset_callback = function(self_auto_projection_size)
      self_auto_projection_size.value=true
      dt.preferences.write(namespace, "checked_auto_projection_size", "bool", true)
    end
}



local pano_pro_combobox_canvas_size = dt.new_widget("combobox")
{
    label = _('output canvas size'),
    tooltip = _('sets the output canvas size'),
    value = dt.preferences.read(namespace, "selected_canvas_size", "integer"),
    changed_callback = function(sel_canvas_size)
      dt.preferences.write(namespace, "selected_canvas_size", "integer", sel_canvas_size.selected)
    end,
    "auto","100%","95%","90%","85%","80%","75%","70%","65%","60%","55%","50%","45%","40%",
    reset_callback = function(self_canvas_size)
      self_canvas_size.value=1
      dt.preferences.write(namespace, "selected_canvas_size", "integer", 1)
    end
}


local pano_pro_slider_exposure = dt.new_widget("slider")
{
  label = _('output exposure value'),
  tooltip = _('Set exposure for ldr mode'),
  hard_min = 0,
  hard_max = 5,
  soft_min = 0,
  soft_max = 5,
  value = dt.preferences.read(namespace, "set_exposure", "float"), --2.0 default
  sensitive = dt.preferences.read(namespace, "visible_exposure", "bool"),
}


local pano_pro_combobox_exposure = dt.new_widget("combobox")
{
    label = _('output exposure'),
    tooltip = _('sets the output exposure'),
    value = dt.preferences.read(namespace, "selected_exposure", "integer"),
    changed_callback = function(sel_exposure)
      dt.preferences.write(namespace, "selected_exposure", "integer", sel_exposure.selected)
      if (sel_exposure.selected == 1) then
          pano_pro_slider_exposure.sensitive=false
          dt.preferences.write(namespace, "visible_exposure", "bool", false)
      else
          pano_pro_slider_exposure.sensitive=true
          dt.preferences.write(namespace, "visible_exposure", "bool", true)
      end
    end,
    "auto","value",
    reset_callback = function(self_exposure)
      self_exposure.value=1
      dt.preferences.write(namespace, "selected_exposure", "integer", 1)
    end
}

local pano_pro_combobox_output_crop = dt.new_widget("combobox")
{
    label = _('output crop'),
    tooltip = _('Sets the crop rectangle.\n\nauto: autocrop panorama\nautohdr: autocrop HDR panorama'),
    value = dt.preferences.read(namespace, "selected_output_crop", "integer"),
    changed_callback = function(sel_output_crop)
      dt.preferences.write(namespace, "selected_output_crop", "integer", sel_output_crop.selected)
    end,
    "auto", "autohdr","no",
    reset_callback = function(self_output_crop)
       self_output_crop.value = 1
       dt.preferences.write(namespace, "selected_output_crop", "integer", 1)
    end
}

local pano_pro_check_button_straighten = dt.new_widget("check_button")
{
    label = _('straightens the panorama'),
    value = dt.preferences.read(namespace, "checked_straighten", "bool"),
    tooltip =_('A common problem when creating a panorama is a curved horizon. This can occur if the panorama was taken with the camera pointing upwards or downwards. Use this option to straightens the panorama'),
    clicked_callback = function(auto_straighten)
    if (auto_straighten.value) then
             dt.preferences.write(namespace, "checked_straighten", "bool", true)
          else
             dt.preferences.write(namespace, "checked_straighten", "bool", false)
          end
    end,
    reset_callback = function(self_straighten)
       self_straighten.value = false
       dt.preferences.write(namespace, "checked_straighten", "bool", false)
    end
}

local pano_pro_check_button_center = dt.new_widget("check_button")
{
    label = _('centers the panorama'),
    value = dt.preferences.read(namespace, "checked_center", "bool"),
    tooltip =_('centers the panorama'),
    clicked_callback = function(center)
    if (center.value) then
             dt.preferences.write(namespace, "checked_center", "bool", true)
          else
             dt.preferences.write(namespace, "checked_center", "bool", false)
          end
    end,
    reset_callback = function(self_center)
       self_center.value = false
       dt.preferences.write(namespace, "checked_center", "bool", false)
    end
}

local pano_pro_separator3 = dt.new_widget("separator")
{
}

local pano_pro_label_output_format= dt.new_widget("label")
{
     label = _('target file'),
     ellipsize = "start",
     halign = "end"
}

local pano_pro_label_line4= dt.new_widget("label")
{
     label = "_______________________________________________",
     ellipsize = "start",
     halign = "end"
}

local pano_pro_combobox_output_format = dt.new_widget("combobox")
{
    label = _('file format'),
    value = dt.preferences.read(namespace, "selected_output_format", "integer"), --2, "TIFF", "JPEG", "PNG","PNM","PBM","PGM","PPM",
    changed_callback = function(sel_output_format)
      dt.preferences.write(namespace, "selected_output_format", "integer", sel_output_format.selected)
    end,
    "TIFF", "JPEG", "PNG",
    reset_callback = function(self_output_format)
       self_output_format.value = dt.preferences.read(namespace, "selected_output_format", "integer")
    end
}

local pano_pro_label_path = dt.new_widget("label")
{
     label = _('directory'),
     ellipsize = "start",
     halign = "start"
}

local pano_pro_file_chooser_button_path_pano = dt.new_widget("file_chooser_button")
{
    title = _('select export path'),  -- The title of the window when choosing a file
    is_directory = true,             -- True if the file chooser button only allows directories to be selecte
    tooltip =_('select the target directory for the panorama image. \nthe filename is created automatically.')
}

local pano_pro_combobox_existing_file = dt.new_widget("combobox")
{
    label = _('on conflict'),
    value = dt.preferences.read(namespace, "selected_overwrite", "integer"), --1,
    changed_callback = function(sel_overwrite)
       dt.preferences.write(namespace, "selected_overwrite", "integer", sel_overwrite.selected)
    end,
    _('create unique filename'),_('overwrite'),
    reset_callback = function(self_overwrite)
       self_overwrite.value = 1
       dt.preferences.write(namespace, "selected_overwrite", "integer", 1)
    end
}

--FUNCTION
local exec_widget
if dt.configuration.running_os ~= "linux" then
  exec_widget = df.executable_path_widget(executable_table)
end

local widget = dt.new_widget("box") {
    orientation = "vertical",
    --project options
    pano_pro_label_project_options,
    pano_pro_label_line1,
    pano_pro_combobox_projection_type,
    pano_pro_check_button_distortion,
    pano_pro_check_button_vignetting,
    pano_pro_check_button_fov,
    pano_pro_slider_fov,
    pano_pro_separator1,
    -- control point options
    pano_pro_label_control_points,
    pano_pro_label_line2,
    pano_pro_combobox_matching_strategy,
    pano_pro_check_button_celeste,
    pano_pro_combobox_celeste_radius,
    pano_pro_combobox_celeste_threshold,
    pano_pro_combobox_ransacmode,
    pano_pro_combobox_ransac_iterations,
    pano_pro_combobox_ransac_distance,
    pano_pro_check_button_fullscale,
    pano_pro_check_button_whole_pano_checking,
    -- panorama options
    pano_pro_separator2,
    pano_pro_label_pano_options,
    pano_pro_label_line3,
    pano_pro_check_button_alignmode,
    pano_pro_check_button_pairwise_optimisation,
    pano_pro_check_button_optimise_photometric_parameters,
    pano_pro_check_button_level_horizon,
    pano_pro_check_button_auto_projection_size,
    pano_pro_combobox_canvas_size,
    pano_pro_combobox_output_crop,
    pano_pro_check_button_straighten,
    pano_pro_check_button_center,
    pano_pro_combobox_exposure,
    pano_pro_slider_exposure,
    -- file options
    pano_pro_separator3,
    pano_pro_label_output_format,
    pano_pro_label_line4,
    pano_pro_combobox_output_format,
    pano_pro_label_path,
    pano_pro_file_chooser_button_path_pano,
    pano_pro_combobox_existing_file,
    exec_widget
}

local function GetFileName(full_path)
  full_path = "x/" .. full_path --workoround because of luas lack with regex in case of single filename
  local filename_with_suffix=full_path:match("^.+[\\/](.+)$")
  return string.sub(filename_with_suffix,1, -5)
end

local function truncate(x)
      return x<0 and math.ceil(x) or math.floor(x)
end

local function replace_comma_to_dot(s)
	return string.gsub(s, "%,", ".")
end

local function file_exists(name)
   local f=io.open(""..name.."","r")
   if f~=nil then
     io.close(f)
     return true
   else
     return false
   end
end

local function remove_tmp_files()
  dt.control.execute("cpfind --clean \""..pano_pro_tmppath.."\\project.pto\"")
  log.msg(log.debug, os.remove(pano_pro_tmppath.."\\project.pto"))
  log.msg(log.debug, "remove temp-files : ", pano_pro_tmppath.."\\project.pto")

  for m in pano_pro_images_to_process:gmatch("[^\"]+%s?[^\"]+") do
    log.msg(log.debug, os.remove(m))
  end
  log.msg(log.debug, "remove temp-files pano_pro_images_to_process: ", pano_pro_images_to_process)

  for m in pano_pro_cmd_enblend_bitmap_input:gmatch("%S+") do
    log.msg(log.debug, os.remove(m:sub(2, -2)))
  end
  log.msg(log.debug, "remove temp-files pano_pro_cmd_enblend_bitmap_input: ", pano_pro_cmd_enblend_bitmap_input)

  pano_pro_counted_images=0
  pano_pro_images_to_process=""
end


local function stop_selection(job)
    job.valid = false
end



local function show_status(storage, image, format, filename,
  number, total, high_quality, extra_data)
     dt.print(_('export TIFF images for panorama ')..tostring(truncate(number)).." / "..tostring(truncate(total)))
end


local function create_pano_image(storage, image_table, extra_data) --finalize

  --log.log_level(log.error)
  log.log_level(log.debug)

  local job = dt.gui.create_job(_('creating panorama image'), true, stop_selection)

  -- delete old projects if exists
  local projectfile=pano_pro_tmppath.."/project.pto"
  if (file_exists(projectfile)) then
    os.remove(pano_pro_tmppath.."/project.pto")
    log.msg(log.info, "old project file project.pto removed")
  end

    -- list of exported images
  local img_set = {}

  -- reset and create image list
  for k,_ in pairs(image_table) do
    table.insert(img_set, k)
  end

  -- use first and last file as basename for output file
  table.sort(img_set, function(a,b) return a.filename < b.filename end);
  local last_file = nil
  local first_file = nil
--  local pano_pro_counted_images=0
  for _, k in ipairs(img_set) do

    if first_file == nil then
      first_file = k
    end

    pano_pro_counted_images=pano_pro_counted_images+1
    last_file = k
  end

  -- create image path and infos
--  pano_pro_counted_images=0
  for _,v in pairs(image_table) do
    pano_pro_images_to_process = pano_pro_images_to_process.."\"" ..v.."\" "
--    pano_pro_counted_images=pano_pro_counted_images+1
--    if (pano_pro_counted_images == 1) then
--      pano_pro_first_imagefile=v
--    else
--      pano_pro_first_last_imagefile=v
--    end
  end

  pano_pro_first_imagefile = first_file.filename
  pano_pro_first_last_imagefile = last_file.filename

  if (pano_pro_counted_images<=1) then
    dt.print(_('ERROR: not enough pictures selected. please select two or more images\nfrom the panorama.'))
    for m in pano_pro_images_to_process:gmatch("%S+") do
      log.msg(log.debug, os.remove(m:sub(2, -2)))
    end
    log.msg(log.error, pano_pro_images_to_process)
    return
  elseif (pano_pro_counted_images>=21) then
    dt.print(_('you have selected more then 20 images. the panorama process could take a very long time! \nhave a nice beake.'))
  end

  local filepath = df.split_filepath(first_file.filename)
  log.msg(log.info, "filepath is ", filepath)

-- directory of the original file
  local dst_path = first_file.path
  log.msg(log.info, "dst_path is ", dst_path)

  local pto_gen = df.check_if_bin_exists("pto_gen")
  log.msg(log.debug, "pto_gen set to ", pto_gen)
  local cpfind = df.check_if_bin_exists("cpfind")
  log.msg(log.debug, "cpfind set to ", cpfind)
  local cpclean = df.check_if_bin_exists("cpclean")
  log.msg(log.debug, "cpclean set to ", cpclean)
  local linefind = df.check_if_bin_exists("linefind")
  log.msg(log.debug, "linefind set to ", linefind)
  local autooptimiser = df.check_if_bin_exists("autooptimiser")
  log.msg(log.debug, "autooptimiser set to ", autooptimiser)
  local pano_modify = df.check_if_bin_exists("pano_modify")
  log.msg(log.debug, "pano_modify set to ", pano_modify)
  local nona = df.check_if_bin_exists("nona")
  log.msg(log.debug, "nona set to ", nona)
  local enblend = df.check_if_bin_exists("enblend")
  log.msg(log.debug, "enblend set to ", enblend)
  local exiftool = df.check_if_bin_exists("exiftool")
  log.msg(log.debug, "exiftool set to ", exiftool)

  if (pto_gen == "") then
    dt.print(_('ERROR: pto_gen not found. please install hugin.'))
    log.msg(log.error, _('pto_gen not found. please install hugin.'))
    for m in pano_pro_images_to_process:gmatch("%S+") do
      log.msg(log.debug, os.remove(m:sub(2, -2)))
    end
    job.valid = false
    return
  elseif (cpfind == "") then
    dt.print(_('ERROR: cpfind not found. please install hugin.'))
    log.msg(log.error,_('cpfind not found. please install hugin.'))
    for m in pano_pro_images_to_process:gmatch("%S+") do
      log.msg(log.debug, os.remove(m:sub(2, -2)))
    end
    job.valid = false
    return
  elseif (cpclean == "") then
    dt.print(_('ERROR: cpclean not found. please install hugin.'))
    log.msg(log.error,_('cpclean not found. please install hugin.'))
    for m in pano_pro_images_to_process:gmatch("%S+") do
      log.msg(log.debug, os.remove(m:sub(2, -2)))
    end
    job.valid = false
    return
  elseif (linefind == "") then
    dt.print(_('ERROR: linefind not found. please install hugin.'))
    log.msg(log.error,_('linefind not found. please install hugin.'))
    for m in pano_pro_images_to_process:gmatch("%S+") do
      log.msg(log.debug, os.remove(m:sub(2, -2)))
    end
    job.valid = false
    return
  elseif (autooptimiser == "") then
    dt.print(_('ERROR: autooptimiser not found. please install hugin.'))
    log.msg(log.error,_('autooptimiser not found. please install hugin.'))
    for m in pano_pro_images_to_process:gmatch("%S+") do
      log.msg(log.debug, os.remove(m:sub(2, -2)))
    end
    job.valid = false
    return
  elseif (pano_modify == "") then
    dt.print(_('ERROR: pano_modify not found. please install hugin.'))
    log.msg(log.error,_('pano_modify not found. please install hugin.'))
    for m in pano_pro_images_to_process:gmatch("%S+") do
      log.msg(log.debug, os.remove(m:sub(2, -2)))
    end
    job.valid = false
    return
  elseif (nona == "") then
    dt.print(_('ERROR: nona not found. please install hugin.'))
    log.msg(log.error,_('nona not found. please install hugin.'))
    for m in pano_pro_images_to_process:gmatch("%S+") do
      log.msg(log.debug, os.remove(m:sub(2, -2)))
    end
    job.valid = false
    return
  elseif (enblend == "") then
    dt.print(_('ERROR: enblend 4.2 not found. please install enblend version 4.2.'))
    log.msg(log.error,_('enblend 4.2 not found. please install enblend version 4.2.'))
    for m in pano_pro_images_to_process:gmatch("%S+") do
      log.msg(log.debug, os.remove(m:sub(2, -2)))
    end
    job.valid = false
    return
  elseif (exiftool == "") then
    dt.print(_('ERROR: exiftool not found. please install exiftool.'))
    log.msg(log.error,_('exiftool not found. please install exiftool.'))
    for m in pano_pro_images_to_process:gmatch("%S+") do
      log.msg(log.debug, os.remove(m:sub(2, -2)))
    end
    job.valid = false
    return
  end

  if dt.configuration.running_os == "linux" then
    local enblendVersionStartCommand='enfuse --version | grep "enfuse 4.2"'
    local enblend_version=dt.control.execute(enblendVersionStartCommand)

    if (enblend_version ~= 0) then
      dt.print(_('ERROR: wrong enblend version found. the plugin works only with enblend 4.2! please install enlendb version 4.2'))
      job.valid = false
      return
    end
  end

  job.percent = 0.1

  -- check path
  if pano_pro_file_chooser_button_path_pano.value == nil then
    pano_pro_cmd_output_path = dst_path
  else
    pano_pro_cmd_output_path = pano_pro_file_chooser_button_path_pano.value
  end

  if (pano_pro_cmd_output_path == nil) then
    dt.print(_('ERROR: no target directory selected'))
    job.valid = false
    return
  else
    -- main program
    -- create and execute pto_gen command
    local ptype=tonumber(pano_pro_combobox_projection_type.selected) -1
    pano_pro_cmd_pto_gen_projection_type="-p "..tostring(ptype)
    if (pano_pro_check_button_distortion.value) then
      pano_pro_cmd_pto_gen_distortion="--distortion"
    else
      pano_pro_cmd_pto_gen_distortion=""
    end

    if (pano_pro_check_button_vignetting.value) then
      pano_pro_cmd_pto_gen_vignetting="--vignetting"
    else
      pano_pro_cmd_pto_gen_vignetting=""
    end

    if (pano_pro_check_button_fov.value) then
      pano_pro_cmd_pto_gen_fov="-f "..pano_pro_slider_fov.value
    else
      pano_pro_cmd_pto_gen_fov=""
    end

    pano_pro_cmd_pto_gen_output="-o \""..pano_pro_tmppath.."/project.pto\""

    pano_pro_pto_genStartCommand="pto_gen "..pano_pro_cmd_pto_gen_projection_type.." "..pano_pro_cmd_pto_gen_distortion.." "..pano_pro_cmd_pto_gen_vignetting.." "..pano_pro_cmd_pto_gen_fov.." "..pano_pro_cmd_pto_gen_output.." "..pano_pro_images_to_process
    log.msg(log.debug, "pano_pro_pto_genStartCommand: " .. pano_pro_pto_genStartCommand)
    dt.print(_('creating panorama project'))
    local result_pto_gen=dt.control.execute(pano_pro_pto_genStartCommand)

    job.percent = 0.2

    if (result_pto_gen == 0) then
    -- create and execute cpfind command

      if (pano_pro_combobox_matching_strategy.selected == 1) then
        pano_pro_pto_cmd_cpfind_matching=""
      elseif (pano_pro_combobox_matching_strategy.selected == 2) then
        pano_pro_pto_cmd_cpfind_matching="--linearmatch"
      elseif (pano_pro_combobox_matching_strategy.selected == 2) then
        pano_pro_pto_cmd_cpfind_matching="--multirow"
      else
        pano_pro_pto_cmd_cpfind_matching=""
      end

      if (pano_pro_check_button_celeste.value) then
        pano_pro_cmd_cpfind_celeste="--celeste"
        pano_pro_cmd_cpfind_celeste_radius="--celesteradius "..pano_pro_combobox_celeste_radius.value
        pano_pro_cmd_cpfind_celeste_threshold="--celestethreshold "..pano_pro_combobox_celeste_threshold.value
      else
        pano_pro_cmd_cpfind_celeste=""
        pano_pro_cmd_cpfind_celeste_radius=""
        pano_pro_cmd_cpfind_celeste_threshold=""
      end

      pano_pro_cmd_cpfind_ransacmode="--ransacmode "..pano_pro_combobox_ransacmode.value
      pano_pro_cmd_cpfind_ransaciter="--ransaciter "..pano_pro_combobox_ransac_iterations.value
      pano_pro_cmd_cpfind_ransacdist="--ransacdist "..pano_pro_combobox_ransac_distance.value

      if (dt.preferences.read(namespace, "keypoint_cache", "bool")) then
        pano_pro_cmd_cpfind_cache="--cache"
      else
        pano_pro_cmd_cpfind_cache=""
      end

      pano_pro_cmd_cpfind_input="\""..pano_pro_tmppath.."/project.pto\""
      pano_pro_cmd_cpfind_output="-o \""..pano_pro_tmppath.."/project.pto\""

      if (pano_pro_check_button_fullscale.value) then
        pano_pro_cmd_cpfind_fullscale="--fullscale"
      else
        pano_pro_cmd_cpfind_fullscale=""
      end

      pano_pro_cpfindStartCommand="cpfind "..pano_pro_pto_cmd_cpfind_matching.." "..pano_pro_cmd_cpfind_celeste.." "..pano_pro_cmd_cpfind_celeste_radius.." "..pano_pro_cmd_cpfind_celeste_threshold.." "..pano_pro_cmd_cpfind_ransacmode.." "..pano_pro_cmd_cpfind_ransaciter.." "..pano_pro_cmd_cpfind_ransacdist.." "..pano_pro_cmd_cpfind_cache.." "..pano_pro_cmd_cpfind_fullscale.." "..pano_pro_cmd_cpfind_output.." "..pano_pro_cmd_cpfind_input
      log.msg(log.debug , "pano_pro_cpfindStartCommand: " .. pano_pro_cpfindStartCommand)
      dt.print(_('creating control points'))
      local result_cpfind=dt.control.execute(pano_pro_cpfindStartCommand)

      job.percent = 0.4

      if (result_cpfind == 0) then
      -- create cpclean command

        if (pano_pro_check_button_whole_pano_checking.value) then
          pano_pro_cmd_cpclean_optimise_whole_panorama="-p"
        else
          pano_pro_cmd_cpclean_optimise_whole_panorama=""
        end

        pano_pro_cmd_cpclean_input="\""..pano_pro_tmppath.."/project.pto\""
        pano_pro_cmd_cpclean_output="-o \""..pano_pro_tmppath.."/project.pto\""
        pano_pro_cpcleanStartCommand="cpclean "..pano_pro_cmd_cpclean_optimise_whole_panorama.." "..pano_pro_cmd_cpclean_output.." "..pano_pro_cmd_cpclean_input
        log.msg(log.debug, "pano_pro_cpcleanStartCommand: " .. pano_pro_cpcleanStartCommand)
        local result_cpclean=dt.control.execute(pano_pro_cpcleanStartCommand)
        log.msg(log.debug, "CPCLEAN RETURN "..result_cpclean)

        job.percent = 0.5

        if (result_cpclean == 0) then
        -- create linefind command
          pano_pro_cmd_linefind_input="\""..pano_pro_tmppath.."/project.pto\""
          pano_pro_cmd_linefind_output="-o \""..pano_pro_tmppath.."/project.pto\""
          pano_pro_linefindStartCommand="linefind "..pano_pro_cmd_linefind_output.." "..pano_pro_cmd_linefind_input
          log.msg(log.debug, "pano_pro_linefindStartCommand: " .. pano_pro_linefindStartCommand)
          dt.print(_('searching lines'))
          local result_linefind=dt.control.execute(pano_pro_linefindStartCommand)
          log.msg(log.debug, "LINEFIND RETURM "..result_linefind)

          job.percent = 0.6

          if (result_linefind == 0) then
          -- create autooptimiser command

            if (pano_pro_check_button_alignmode.value) then
              pano_pro_cmd_autooptimiser_auto_align_mode="-a"
            else
              pano_pro_cmd_autooptimiser_auto_align_mode=""
            end

            if (pano_pro_check_button_pairwise_optimisation.value) then
              pano_pro_cmd_autooptimiser_pairwise_optimisation="-p"
            else
              pano_pro_cmd_autooptimiser_pairwise_optimisation=""
            end

            if (pano_pro_check_button_optimise_photometric_parameters.value) then
              pano_pro_cmd_autooptimiser_photometric_parameters="-m"
            else
              pano_pro_cmd_autooptimiser_photometric_parameters=""
            end

            if (pano_pro_check_button_level_horizon.value) then
              pano_pro_cmd_autooptimiser_level_horizon="-l"
            else
              pano_pro_cmd_autooptimiser_level_horizon=""
            end

            if (pano_pro_check_button_auto_projection_size.value) then
              pano_pro_cmd_autooptimiser_auto_projection="-s"
            else
              pano_pro_cmd_autooptimiser_auto_projection=""
            end

            pano_pro_cmd_autooptimiser_input="\""..pano_pro_tmppath.."/project.pto\""
            pano_pro_cmd_autooptimiser_output="-o \""..pano_pro_tmppath.."/project.pto\""
            pano_pro_autooptimiserStartCommand="autooptimiser "..pano_pro_cmd_autooptimiser_auto_align_mode.." "..pano_pro_cmd_autooptimiser_pairwise_optimisation.." "..pano_pro_cmd_autooptimiser_photometric_parameters.." "..pano_pro_cmd_autooptimiser_level_horizon.." "..pano_pro_cmd_autooptimiser_auto_projection.." "..pano_pro_cmd_autooptimiser_output.." "..pano_pro_cmd_autooptimiser_input
            log.msg(log.debug, "pano_pro_autooptimiserStartCommand: " .. pano_pro_autooptimiserStartCommand)
            dt.print(_('optimizing image positions'))
            local result_autooptimiser=dt.control.execute(pano_pro_autooptimiserStartCommand)
            log.msg(log.debug, "AUTOOPTIMIZER RETURN "..result_autooptimiser)

            job.percent = 0.7

            if (result_autooptimiser == 0) then
            -- create pano_modify command
              pano_pro_cmd_pano_modify_canvas="--canvas="..pano_pro_combobox_canvas_size.value

              if (pano_pro_combobox_output_crop.selected == 3) then
                pano_pro_cmd_pano_modify_crop=""
              else
                pano_pro_cmd_pano_modify_crop="--crop="..pano_pro_combobox_output_crop.value
              end

              if (pano_pro_check_button_straighten.value) then
                pano_pro_cmd_pano_modify_straighten="--straighten"
              else
                pano_pro_cmd_pano_modify_straighten=""
              end

              if (pano_pro_check_button_center.value) then
                pano_pro_cmd_pano_modify_center="--center"
              else
                pano_pro_cmd_pano_modify_center=""
              end

              if (pano_pro_combobox_exposure.selected == 1) then
                pano_pro_cmd_pano_modify_exposure="--output-exposure=auto"
              else
                pano_pro_cmd_pano_modify_exposure="--output-exposure="..(replace_comma_to_dot(pano_pro_slider_exposure.value))
              end

              if (pano_pro_combobox_output_format.selected == 1) then
                pano_pro_cmd_pano_modify_output_format="--ldr-file=TIF"
                pano_pro_cmd_pano_modify_output_compression="--ldr-compression="..dt.preferences.read(namespace, "compression_tiff", "enum")
              elseif (pano_pro_combobox_output_format.selected == 2) then
                pano_pro_cmd_pano_modify_output_format="--ldr-file=JPG"
                pano_pro_cmd_pano_modify_output_compression="--ldr-compression="..truncate(dt.preferences.read(namespace, "compression_jpeg", "integer"))
              elseif (pano_pro_combobox_output_format.selected == 3) then
                pano_pro_cmd_pano_modify_output_format="--ldr-file=PNG"
              end

              pano_pro_cmd_pano_modify_input="\""..pano_pro_tmppath.."/project.pto\""
              pano_pro_cmd_pano_modify_output="-o \""..pano_pro_tmppath.."/project.pto\""
              pano_pro_pano_modifyStartCommand="pano_modify --blender=ENBLEND "..pano_pro_cmd_pano_modify_canvas.." "..pano_pro_cmd_pano_modify_crop.." "..pano_pro_cmd_pano_modify_straighten.." "..pano_pro_cmd_pano_modify_center.." "..pano_pro_cmd_pano_modify_exposure.." "..pano_pro_cmd_pano_modify_output.." "..pano_pro_cmd_pano_modify_input
              log.msg(log.debug, "pano_pro_pano_modifyStartCommand: " .. pano_pro_pano_modifyStartCommand)
              local result_pano_modify=dt.control.execute(pano_pro_pano_modifyStartCommand)

              job.percent = 0.8

              if (result_pano_modify == 0) then
              -- create nona command
                pano_pro_cmd_nona_prj_input="\""..pano_pro_tmppath.."/project.pto\""
                pano_pro_cmd_nona_bitmap_output="\""..pano_pro_tmppath.."/nona\""

                if (dt.preferences.read(namespace, "nona_use_gpu", "bool")) then
                  pano_pro_cmd_nona_use_gpu="-g"
                else
                  pano_pro_cmd_nona_use_gpu=""
                end

                if (pano_pro_combobox_exposure.selected == 1) then
                  pano_pro_cmd_nona_exposure=""
                else
                  pano_pro_cmd_nona_exposure="-e "..(replace_comma_to_dot(pano_pro_slider_exposure.value))
                end

                pano_pro_nonaStartCommand="nona -v -m TIFF_m "..pano_pro_cmd_nona_use_gpu.." "..pano_pro_cmd_nona_exposure.." -o "..pano_pro_cmd_nona_bitmap_output.." "..pano_pro_cmd_nona_prj_input
                log.msg(log.debug, "pano_pro_nonaStartCommand: " .. pano_pro_nonaStartCommand)
                dt.print(_('merging images'))
                local result_nona=dt.control.execute(pano_pro_nonaStartCommand)
                log.msg(log.debug, "NONA RETURN "..result_nona)

                job.percent = 0.9

                if (result_nona == 0) then
                -- create enblend command
                  local tiff_index=0
                  local cmd_suffix_output_format
                  local outputindex
                  for x=1, pano_pro_counted_images do
                    local nona_filename_index
                    if (tiff_index <=9) then
                      nona_filename_index="000"..(tostring(tiff_index))
                    elseif (tiff_index <=99) then
                      nona_filename_index="00"..(tostring(tiff_index))
                    elseif (tiff_index <=999) then
                      nona_filename_index="0"..(tostring(tiff_index))
                    end

                    pano_pro_cmd_enblend_bitmap_input = pano_pro_cmd_enblend_bitmap_input.."\""..pano_pro_tmppath.."/nona"..nona_filename_index..".tif\" "
                    tiff_index=tiff_index+1
                  end

                  if (pano_pro_combobox_output_format.selected == 1) then
                    cmd_suffix_output_format="tif"
                    pano_pro_cmd_enblend_output_compression="--compression "..dt.preferences.read(namespace, "compression_tiff", "enum")
                    pano_pro_cmd_enblend_output_depth="-d "..dt.preferences.read(namespace, "image_color_depth", "enum")
                  elseif (pano_pro_combobox_output_format.selected == 2) then
                    cmd_suffix_output_format="jpg"
                    pano_pro_cmd_enblend_output_compression="--compression "..dt.preferences.read(namespace, "compression_jpeg", "integer")
                    pano_pro_cmd_enblend_output_depth=""
                  elseif (pano_pro_combobox_output_format.selected == 3) then
                    cmd_suffix_output_format="png"
                    pano_pro_cmd_enblend_output_compression=""
                    pano_pro_cmd_enblend_output_depth=""
                  end

                  if pano_pro_file_chooser_button_path_pano.value == nil then
                    pano_pro_cmd_output_path =  dst_path
                  else
                    pano_pro_cmd_output_path = pano_pro_file_chooser_button_path_pano.value
                  end

                  if(pano_pro_combobox_existing_file.selected == 1) then
                    pano_pro_path_with_filename=pano_pro_cmd_output_path.."/"..(GetFileName(pano_pro_first_imagefile)).."-"..(GetFileName(pano_pro_first_last_imagefile)).."."..cmd_suffix_output_format

                    while (file_exists(pano_pro_path_with_filename)) do
                      pano_pro_cmd_output_image_index=pano_pro_cmd_output_image_index+1

                      if pano_pro_cmd_output_image_index <= 9 then
                        outputindex="0"..tostring(pano_pro_cmd_output_image_index)
                      else
                        outputindex=tostring(pano_pro_cmd_output_image_index)
                      end
                      pano_pro_path_with_filename=pano_pro_cmd_output_path.."/"..(GetFileName(pano_pro_first_imagefile)).."-"..(GetFileName(pano_pro_first_last_imagefile)).."_"..outputindex.."."..cmd_suffix_output_format
                    end

                    pano_pro_cmd_output_image_index=0
                    pano_pro_cmd_enblend_bitmap_output="\""..pano_pro_path_with_filename.."\""
                  else
                    pano_pro_path_with_filename=pano_pro_cmd_output_path.."/"..(GetFileName(pano_pro_first_imagefile)).."-"..(GetFileName(pano_pro_first_last_imagefile)).."."..cmd_suffix_output_format
                    pano_pro_cmd_enblend_bitmap_output="\""..pano_pro_path_with_filename.."\""
                  end

                  log.msg(log.debug, "pano_pro_cmd_enblend_bitmap_output: " .. pano_pro_cmd_enblend_bitmap_output)
                  if df.check_if_file_exists(pano_pro_cmd_enblend_bitmap_output) then
                    pano_pro_cmd_enblend_bitmap_output = df.create_unique_filename(pano_pro_cmd_enblend_bitmap_output)
                    log.msg(log.info, "pano_pro_cmd_enblend_bitmap_output updated to ", pano_pro_cmd_enblend_bitmap_output)
                  end

                  if (dt.preferences.read(namespace, "nona_use_gpu", "bool")) then
                    pano_pro_cmd_enblend_use_gpu="--gpu"
                  else
                    pano_pro_cmd_enblend_use_gpu=""
                  end

                  pano_pro_enblendStartCommand="enblend -v "..pano_pro_cmd_enblend_use_gpu.." "..pano_pro_cmd_enblend_output_compression.." "..pano_pro_cmd_enblend_output_depth.." -o "..pano_pro_cmd_enblend_bitmap_output.." "..pano_pro_cmd_enblend_bitmap_input
                  log.msg(log.debug, "pano_pro_enblendStartCommand: " .. pano_pro_enblendStartCommand)
                  local result_enblend=dt.control.execute(pano_pro_enblendStartCommand)
                  log.msg(log.debug, "ENBLEND RETURN "..result_enblend)

                  local exifStartCommand
                  local resultexif
                  if ((result_enblend == 0) and (dt.preferences.read(namespace, "exiftool_copy_tags", "bool")) and (dt.preferences.read(namespace, "add_image_to_db", "bool"))) then
                  --attach exif data and add to db
                    exifStartCommand="exiftool -TagsFromFile \""..pano_pro_first_imagefile.."\" -exif:all --subifd:all -overwrite_original "..pano_pro_cmd_enblend_bitmap_output..""
                    exifStartCommand="exiftool -TagsFromFile \""..pano_pro_first_imagefile.."\" -exif:all --subifd:all -overwrite_original "..pano_pro_cmd_enblend_bitmap_output..""
                    log.msg(log.debug, "exifStartCommand: " .. exifStartCommand)
                    resultexif=dt.control.execute(exifStartCommand)
                    local image = dt.database.import(pano_pro_path_with_filename)
                    log.msg(log.debug, "pano_pro_path_with_filename: " .. pano_pro_path_with_filename)
                    remove_tmp_files()
                    job.valid = false
                    dt.print(_('process successfully completed'))
                  elseif ((result_enblend == 0) and (dt.preferences.read(namespace, "exiftool_copy_tags", "bool")) and (dt.preferences.read(namespace, "add_image_to_db", "bool") == false)) then
                  --attach exif data only
                    exifStartCommand="exiftool -TagsFromFile \""..pano_pro_first_imagefile.."\" -exif:all --subifd:all -overwrite_original "..pano_pro_cmd_enblend_bitmap_output..""
                    log.msg(log.debug, "exifStartCommand: " .. exifStartCommand)
                    resultexif=dt.control.execute(exifStartCommand)
                    log.msg(log.debug, "EXIFTOOL RETURN "..resultexif)

                    if (resultexif == 0) then
                      remove_tmp_files()
                      job.valid = false
                      dt.print(_('process successfully completed'))
                    else
                      remove_tmp_files()
                      job.valid = false
                      dt.print(_('ERROR: exiftool doesn\'t work. for more informations see terminal output'))
                    end

                  elseif ((result_enblend == 0) and (dt.preferences.read(namespace, "exiftool_copy_tags", "bool") == false) and (dt.preferences.read(namespace, "add_image_to_db", "bool"))) then
                  -- add to db only
                    local image = dt.database.import(pano_pro_path_with_filename)
                    log.msg(log.debug, "pano_pro_path_with_filename: " .. pano_pro_path_with_filename)
                    remove_tmp_files()
                    job.valid = false
                    dt.print(_('process successfully completed'))
                  elseif ((result_enblend == 0) and (dt.preferences.read(namespace, "exiftool_copy_tags", "bool") == false) and (dt.preferences.read(namespace, "add_image_to_db", "bool") == false)) then
                  -- exit without exif tags and add to db
                    remove_tmp_files()
                    job.valid = false
                    dt.print(_('process successfully completed'))
                  elseif (result_enblend == 256) then
                    remove_tmp_files()
                    job.valid = false
                    dt.print(_('ERROR: nona could not create any images. \nthe overlapping of the images may not be sufficient, \nor insufficient control points could be created. \nfor more informations see terminal output'))
                  else
                    -- enblend error
                    dt.print(_('ERROR: enblend doesn\'t work. for more informations see terminal output'))
                    remove_tmp_files()
                    job.valid = false
                  end

                else
                  dt.print(_('ERROR: nona doesn\'t work. for more informations see terminal output'))
                  remove_tmp_files()
                  job.valid = false
                end

              else
                dt.print(_('ERROR: pano_modify doesn\'t work. for more informations see terminal output'))
                remove_tmp_files()
                job.valid = false
              end

            else
              dt.print(_('ERROR: autooptimiser doesn\'t work. for more informations see terminal output'))
              remove_tmp_files()
              job.valid = false
            end

          else
            dt.print(_('ERROR: linefind doesn\'t work. for more informations see terminal output'))
            remove_tmp_files()
            job.valid = false
          end

        else
          dt.print(_('ERROR: cpclean doesn\'t work. for more informations see terminal output'))
          remove_tmp_files()
          job.valid = false
        end
      else
        dt.print(_('ERROR: cpfind doesn\'t work. for more informations see terminal output'))
        remove_tmp_files()
        job.valid = false
      end

    else
      dt.print(_('ERROR: pto_gen doesn\'t work. for more informations see terminal output'))
      remove_tmp_files()
      job.valid = false
    end
  end
end


-- limit export to tiff
local function support_format(storage, format)
  local fmt = string.lower(format.name)
  if string.match(fmt,"tiff") == nil then
    return false
  else
    return true
  end
end



-- REGISTER

dt.register_storage(namespace, _('panorama image'), show_status, create_pano_image, support_format, nil, widget)



dt.preferences.register(namespace,
                        "keypoint_cache",                                          -- name
                        "bool",                                                   -- type
                        _('pano pro: keypoint cache'),                       -- label
                        _('use cache by keypoint calculation'),                  -- tooltip
                        true)                                                    -- default

dt.preferences.register(namespace,
                        "nona_use_gpu",                                          -- name
                        "bool",                                                   -- type
                        _('pano pro: use GPU for remaping'),                       -- label
                        _('set the GPU remapping for image align'),                  -- tooltip
                        false)                                                    -- default

dt.preferences.register(namespace,
                        "exiftool_copy_tags",                                     -- name
                        "bool",                                                   -- type
                        _('pano pro: copy exif tags'),                             -- label
                        _('copy the exif tags from the first image to the target'),  -- tooltip
                        true)                                                     -- default

dt.preferences.register(namespace,
                        "add_image_to_db",                                        -- name
                        "bool",                                                   -- type
                        _('pano pro: add panorama image to database'),                -- label
                        _('add the panorama image to the darktable database'),          -- tooltip
                        false)                                                    -- default

dt.preferences.register(namespace,
                        "compression_jpeg",                        -- name
                        "integer",                                 -- type
                        _('pano pro: JPEG compression'),            -- label
                        _('set the compression for JPEG files'),      -- tooltip
                        98,                                        -- default
                        50,                                        -- min
                        100)                                       -- max

dt.preferences.register(namespace,
                        "compression_tiff",                        -- name
                        "enum",                                    -- type
                        _('pano pro: TIFF compression'),            -- label
                        _('set the compression type for tiff files'), -- tooltip
                        "LZW",                                     -- default
                        "NONE", "DEFLATE","LZW","PACKBITS")        -- values

dt.preferences.register(namespace,
                        "image_color_depth",                       -- name
                        "enum",                                    -- type
                        _('pano pro: image color depth (bit)'),     -- label
                        _('set the output color depth'),              -- tooltip
                        "16",                                      -- default
                        "8","16")                                  -- values
